/*
 * Copyright (c) 2017, 2019, Oracle and/or its affiliates. All rights reserved.
 *
 * This software is dual-licensed to you under the MIT License (MIT) and
 * the Universal Permissive License (UPL). See the LICENSE file in the root
 * directory for license terms. You may choose either license, or both.
 */

#include <stdlib.h>
#include <stdio.h>

#include "iotcs.h"
#include "iotcs/iotcs_private.h"
#include "iotcs_port_crypto.h"
#include "iotcs_port_system.h"
#include "protocol/protocol_request.h"
#include "util/util.h"
#include "util/util_memory.h"
#include "util/util_thread_private.h"
#include "trusted_assets_manager/iotcs_tam.h"
#include "iotcs_port_ssl.h"
#include "scs/storage_dispatcher_private.h"
#include "protocol/protocol_request.h"

#include "log/log.h"
#define IOTCSP_MODULE_LOG_CHANNEL LOG_CHANNEL_CORE
#include "log/log_template.h"

#define MAX_HOST_LEN 256
#define MAX_BODY_LEN 256
#define TMP_BUFFER_LEN 256
#define MAX_API_LEN 256
#define DEFAULT_SCS_PORT 443
#define CHUNK_OVERHEAD 8 // %04X + /r/n + /r/n

#ifndef IOTCS_RESPONSE_BUFFER_LENGTH
#define IOTCS_RESPONSE_BUFFER_LENGTH 4096
#endif

#define EXPECTED_RESPONSE_CODE_COUNT 8
#define USEFUL_DIGEST_LEN 21

static char g_storage_response_buffer[IOTCS_RESPONSE_BUFFER_LENGTH];
static char g_storage_host[MAX_HOST_LEN];
int g_storage_initialized = IOTCS_FALSE;
static iotcs_bool g_storage_reconfig_required = IOTCS_TRUE;
static int g_storage_port = 0;
static char g_storage_rest_api[MAX_API_LEN];
static char g_storage_auth_token[UTIL_TOKEN_STRING_MAX_LENGTH];
char *g_storage_container_url = NULL;
static iotcs_key_value *g_storage_auth_header = NULL;
static int g_storage_auth_header_len = 0;

#define IS_CLASSIC_OBJECT_STORAGE   g_storage_auth_token[0]
#define IOT_PROVISION_REST          "/iot/api/v2/provisioner/storage"
#define APP_JSON_HEADER             "application/json"

iotcs_result util_file_get_md5(char *output_buffer, const char *path, char* tmp_buffer, int tmp_buffer_len);

//
// For Object Storage, response looks like
// {
//   "storageUrl": "https://{server}/v1/c/{compartment}/b/{bucket}/o/{object}",
//   "headers": {
//        "Authorization": "Signature keyId=\"ocid1.tenan...,algorithm=...,headers=...,signature=...\"",
//        "date":"some date in SimpleDateFormat: EEE, dd MMM yyyy HH:mm:ss zzz",
//        "host":"URL of host",
//        "(request-target)": "get /v1/iotdev/Test"
//   }
// }
//
// For Object Storage Classic, response looks like
// {
//    "storageContainerUrl":"https://{instance}/v1/{account}/{container}",
//    "authToken":"AUTH_tk9e18bb3eba2eb0c5d16847218a12a36f"
// }
//

/*Should be guarded (or called from guarded methods) by UTIL_LOCK_STORAGE/UTIL_UNLOCK_STORAGE */
static iotcs_result scs_parse_auth_response(const char* body) {
    json_tokens_t tokens, header, params_item_tok;
    json_object_iterator_t iter;
    char* value_str = "";
    int value_len = 0, i;
    char* tmp = NULL;
    char* next = NULL;
    char* server_scheme = NULL;
    char* server_uri = NULL;
    char* server_name = NULL;
    char* server_port = NULL;
    iotcs_result rv = IOTCS_RESULT_FAIL;
    iotcs_key_value *params;

    GOTO_ERR_MSG(json_parse(body, strlen(body), &tokens) != IOTCS_RESULT_OK,
            "Cannot parse SCS auth JSON.");

    if (json_get_string_value_by_key(&tokens, "storageContainerUrl", &value_str,
            &value_len) != IOTCS_RESULT_OK) { // If it is not Object Storage Classic
        GOTO_ERR_MSG(json_get_string_value_by_key(&tokens, "storageUrl", &value_str,
                &value_len) != IOTCS_RESULT_OK, "Cannot find storageUrl in SCS auth JSON");
    }

    GOTO_ERR_OOM(rv, tmp = util_safe_strncpy(value_str, value_len)); /*now storageContainerUrl null terminated*/

    GOTO_ERR_OOM(rv, g_storage_container_url = util_safe_strncpy(value_str, value_len));
    LOG_INFO("Storage Container Url %s", g_storage_container_url);
    next = tmp;

    /*
     * parse and check the storageContainerUrl
     * example https://host[:port]/v1/Storage-a210401/container 
     */

    GOTO_ERR_MSG(NULL == (server_scheme = util_get_next_token(&next, '/')),
            "Bad storageContainerUrl format. There is no server scheme.");
    GOTO_ERR_MSG((strcmp("https:", server_scheme) != 0) && (strcmp("http:", server_scheme) != 0),
            "Bad storageContainerUrl format. Unexpected server scheme.");
    /*empty string is expected because of //*/
    GOTO_ERR_MSG(NULL == util_get_next_token(&next, '/'),
            "Bad storageContainerUrl format.");
    /*host[:port]*/
    GOTO_ERR_MSG(NULL == (server_uri = util_get_next_token(&next, '/')),
            "Bad storageContainerUrl format. There is no storage cloud service server value in the uri.");

    /*store storageContainerUrl to global variable before checking because tmp str will be changed*/
    GOTO_ERR_MSG(0 > util_safe_snprintf(g_storage_rest_api, MAX_API_LEN, "%s", next),
            "scs_parse_auth_response method failed: storageContainerUrl size is longer then expected.");

    /*api version part - useless information*/
    GOTO_ERR_MSG(NULL == util_get_next_token(&next, '/'),
            "Bad storageContainerUrl format. There is no api version part in the uri.");
    /*storage service identity*/
    GOTO_ERR_MSG(NULL == util_get_next_token(&next, '/'),
            "Bad storageContainerUrl format. There is no storage service identity in the uri.");
    /*storage service container*/
    GOTO_ERR_MSG(NULL == util_get_next_token(&next, '/'),
            "Bad storageContainerUrl format. There is no storage service container in the uri.");

    /*if we are here then URi is correct*/

    /*parse host name and port*/
    next = server_uri;
    GOTO_ERR_MSG(NULL == (server_name = util_get_next_token(&next, ':')),
            "Bad storageContainerUrl format. There is no server name.");

    /*store scs_host to global variable before checking because tmp str will be changed*/
    GOTO_ERR_MSG(0 > util_safe_snprintf(g_storage_host, MAX_HOST_LEN, "%s", server_name),
            "scs_parse_auth_response method failed: storage server name size is longer then expected.");

    if (NULL != (server_port = util_get_next_token(&next, ':'))) {
        /*server port was stored to server_uri*/
        g_storage_port = atoi(server_port);
    } else {
        g_storage_port = DEFAULT_SCS_PORT;
    }
    LOG_INFO("Storage server %s port %d", g_storage_host, g_storage_port);

    tmp = NULL;
    next = NULL;

    /*take and save authToken*/
    if (json_get_string_value_by_key(&tokens, "authToken", &value_str,
            &value_len) != IOTCS_RESULT_OK) { // If it is not Object Storage Classic
        GOTO_ERR_MSG(json_get_object_by_key(&tokens, "headers", &header) != IOTCS_RESULT_OK,
            "Cannot find authToken in SCS auth JSON");
        if (g_storage_auth_header_len) {
            for (i = 0; i < g_storage_auth_header_len; i++) {
                util_free((void*)g_storage_auth_header[i].key);
                util_free((void*)g_storage_auth_header[i].value);
            }
            util_free((void*)g_storage_auth_header);
            g_storage_auth_header = NULL;
            g_storage_auth_header_len = 0;
        }

        g_storage_auth_header_len = json_child_items_count(&header);
        if (g_storage_auth_header_len != 0) {

            CHECK_OOM_MSG(g_storage_auth_header = util_malloc(g_storage_auth_header_len * sizeof (iotcs_key_value)),
                            "Out of memory exception while parse auth response for Object Storage");
            GOTO_ERR_NO_OK(rv, json_get_object_iterator(&header, &iter));

            params = g_storage_auth_header;
            char * tmp_buffer;
            while (json_object_iterator_has_next(&iter)) {
                GOTO_ERR_NO_OK(rv, json_object_iterator_next(&iter, &params_item_tok));

                GOTO_ERR_NO_OK(rv, json_get_string_and_move(&params_item_tok, &value_str, &value_len));
                CHECK_OOM_MSG(params->key = util_safe_strncpy(value_str, value_len),
                            "Out of memory exception while parse auth response for Object Storage");

                GOTO_ERR_NO_OK(rv, json_get_string_and_move(&params_item_tok, &value_str, &value_len));
                CHECK_OOM_MSG(tmp_buffer = util_safe_strncpy(value_str, value_len),
                            "Out of memory exception while parse auth response for Object Storage");
                int i, k;
                for (i = 0, k = 0; i < value_len; i++) {
                    if (tmp_buffer[i] == '\\') {
                        if (i && tmp_buffer[i - 1] != '\\') {
                            k++;
                        } else {
                            tmp_buffer[i - k] = tmp_buffer[i];
                        }
                    } else {
                        tmp_buffer[i - k] = tmp_buffer[i];
                    }
                }
                tmp_buffer[i - k] = '\0';
                params->value = tmp_buffer;
                params++;

            }
        }
        g_storage_auth_token[0] = '\0';
    } else { // Classic
        GOTO_ERR_OOM(rv, tmp = util_safe_strncpy(value_str, value_len)); /*now authToken null terminated*/
        /*store authToken to global variable*/
        GOTO_ERR_MSG(0 > util_safe_snprintf(g_storage_auth_token, UTIL_TOKEN_STRING_MAX_LENGTH, "%s", tmp),
                "scs_parse_auth_response method failed: authToken size is longer then expected.");
        if (g_storage_auth_header_len) {
            for (i = 0; i < g_storage_auth_header_len; i++) {
                util_free((void*)g_storage_auth_header[i].key);
                util_free((void*)g_storage_auth_header[i].value);
            }
        }
        util_free((void*)g_storage_auth_header);
        g_storage_auth_header = NULL;
        g_storage_auth_header_len = 0;
    }


    /*clean up allocated memory*/
    util_free(tmp);
    tmp = NULL;

    return IOTCS_RESULT_OK;

error:
    util_free(tmp);
    return rv;
}

/*Should be guarded (or called from guarded methods) by UTIL_LOCK_STORAGE/UTIL_UNLOCK_STORAGE */

/*Should be guarded (or called from guarded methods) by UTIL_LOCK_LL/UTIL_UNLOCK_LL */
static iotcs_result scs_provision_request_unsafe(char* object_name, char* method) {
    iotcs_result result = IOTCS_RESULT_FAIL;
    protocol_response response;
    char req_body[MAX_BODY_LEN];
    req_body[0]='\0';

    protocol_request request = {
        .body = req_body,
        .url = IOT_PROVISION_REST,
        .method = PROTOCOL_REQUEST_METHOD_POST,
        .headers =
        {
            .connection = "close",
            .accept = APP_JSON_HEADER,
            .content_type = APP_JSON_HEADER,
            .host = tam_get_server_host(),
            .x_endpointId = tam_get_endpoint_id()
        }
    };
    if (method) { // Generate body for Object Storage
        if (util_safe_snprintf(req_body, MAX_BODY_LEN, "{\"object\": \"%s\", \"method\": \"%s\"}",
                               object_name, method) <= 0) {
            result = IOTCS_RESULT_FAIL;
        }
    }

    result = cl_proceed_request(request, &response);
    if (result == IOTCS_RESULT_CANNOT_AUTHORIZE) {
        if (cl_get_token() == IOTCS_RESULT_OK) {
            result = cl_proceed_request(request, &response);
        }
    }

    if (result == IOTCS_RESULT_OK) {
        if (response.status_code == PROTOCOL_RESPONSE_CODE_OK ||
            response.status_code == PROTOCOL_RESPONSE_CODE_CREATED) {
            if (response.body == NULL) {
                result = IOTCS_RESULT_FAIL;
            } else {
                result = scs_parse_auth_response(response.body);
            }
        } else {
            protocol_request old_request = {
                .body = "",
                .url = IOT_PROVISION_REST,
                .method = PROTOCOL_REQUEST_METHOD_GET,
                .headers =
                {
                    .connection = "close",
                    .accept = APP_JSON_HEADER,
                    .host = tam_get_server_host(),
                    .x_endpointId = tam_get_endpoint_id()
                }
            };
            req_body[0]='\0';
            result = cl_proceed_request(old_request, &response);
            if (result == IOTCS_RESULT_CANNOT_AUTHORIZE) {
                if (cl_get_token() == IOTCS_RESULT_OK) {
                    result = cl_proceed_request(old_request, &response);
                }
            }

            if (result == IOTCS_RESULT_OK) {
                if (response.status_code == PROTOCOL_RESPONSE_CODE_OK ||
                    response.status_code == PROTOCOL_RESPONSE_CODE_CREATED) {
                    if (response.body == NULL) {
                        result = IOTCS_RESULT_FAIL;
                    } else {
                        result = scs_parse_auth_response(response.body);
                    }
                } else {
                    result = IOTCS_RESULT_FAIL;
                }
            }
        }
    }

    return result;
}

/*Should be guarded (or called from guarded methods) by UTIL_LOCK_STORAGE/UTIL_UNLOCK_STORAGE */
/*Should be guarded (or called from guarded methods) by UTIL_LOCK_LL/UTIL_UNLOCK_LL */
static iotcs_result cl_internal_storage_reconfig_unsafe(char* object_name, char* method);


char *build_object_path(struct sd_storage_object_t *storage_object) {
    char *object_path = NULL;
    int len = 0;

    if (storage_object->endpoint_id && !DISABLE_STORAGE_OBJECT_PREFIX) {
        len = strlen(storage_object->endpoint_id) + strlen(storage_object->name) + 2;
        object_path = (char*)malloc(len * sizeof(char));
        util_safe_snprintf(object_path, len, "%s/%s", storage_object->endpoint_id, storage_object->name);
    } else {
        len = strlen(storage_object->name) + 2;
        object_path = (char*)malloc(len * sizeof(char));
        util_safe_snprintf(object_path, len, "%s", storage_object->name);
    }

    return object_path;
}

iotcs_result storage_authentication(char* object_name, char* method) {
    UTIL_LOCK_LL();
    if (IOTCS_RESULT_OK != cl_internal_storage_reconfig_unsafe(object_name, method)) {
        UTIL_UNLOCK_LL();
        return IOTCS_RESULT_CANNOT_AUTHORIZE;
    } else {
        UTIL_UNLOCK_LL();
        return IOTCS_RESULT_OK;
    }
    return IOTCS_RESULT_OK;
}

static iotcs_result cl_internal_storage_reconfig_unsafe(char* object_name, char* method) {
    LOG_INFOS(">>>>>>>>>>>>>>>>>>>>> SCS Reconfiguration started");

    char old_storage_host[MAX_HOST_LEN];
    int old_port = g_storage_port;

    memcpy(old_storage_host, g_storage_host, strlen(g_storage_host));

    if (IOTCS_RESULT_OK != scs_provision_request_unsafe(object_name, method)) {
        LOG_ERRS("scs_provision_request_unsafe method failed");
        return IOTCS_RESULT_FAIL;
    }

    if (old_port != g_storage_port || strcmp(old_storage_host, g_storage_host) != 0) {
        iotcs_port_storage_ssl_finalize();

        iotcs_bool is_ssl = (strncmp("https", g_storage_container_url, strlen("https")) == 0) ? IOTCS_TRUE : IOTCS_FALSE;
        if (IOTCS_RESULT_OK != iotcs_port_storage_ssl_init(g_storage_host, g_storage_port, is_ssl)) {
            LOG_ERRS("iotcs_port_storage_ssl_init method failed");
            return IOTCS_RESULT_FAIL;
        }
    }

    g_storage_reconfig_required = IOTCS_FALSE;
    LOG_INFOS(">>>>>>>>>>>>>>>>>>>>> SCS Reconfiguration finished");
    return IOTCS_RESULT_OK;
}


/*Should be guarded (or called from guarded methods) by UTIL_LOCK_LL/UTIL_UNLOCK_LL */
iotcs_result cl_internal_storage_init() {
    iotcs_result result = IOTCS_RESULT_OK;

    UTIL_LOCK_STORAGE();
    /*Configure async dispatcher*/
    GOTO_ERR_MSG((result = cl_internal_storage_async_dispatcher_init()) != IOTCS_RESULT_OK,
            "cl_internal_storage_async_dispatcher_init method was failed.");

    g_storage_reconfig_required = IOTCS_FALSE;
    g_storage_initialized = IOTCS_TRUE;
error:
    UTIL_UNLOCK_STORAGE();
    return result;
}

/*safe*/
void cl_internal_storage_finalize() {
    UTIL_LOCK_STORAGE();
    if (g_storage_initialized) {
        cl_internal_storage_async_dispatcher_finalize();
        iotcs_port_storage_ssl_finalize();
        g_storage_initialized = IOTCS_FALSE;
        g_storage_port = 0;
        util_free(g_storage_container_url);
        g_storage_container_url = NULL;
    }
    UTIL_UNLOCK_STORAGE();
}

/*Should be guarded (or called from guarded methods) by UTIL_LOCK_STORAGE/UTIL_UNLOCK_STORAGE*/
iotcs_result cl_internal_send_upload_request_unsafe(iotcs_storage_request* request, protocol_response* response) {
    iotcs_result result = IOTCS_RESULT_FAIL;
    int written_bytes = 0;
    int cx = 0, i = 0;
    int64_t bytes_total = 0;
    FILE* hInput = NULL;
    char *tmp = NULL;
    char calc_hash[2 * IOTCS_MD5_DIGEST_LENGTH + 1], hash[2 * IOTCS_MD5_DIGEST_LENGTH + 1];
    char tmp_hash_buffer[TMP_BUFFER_LEN];
    struct sd_storage_object_t *storage_object = (struct sd_storage_object_t *)request->storage->object;

    /*check cancellation mark*/
    if (cl_storage_progress_cb_h(request->storage, storage_object->progress_state, bytes_total, NULL) == IOTCS_TRUE) {
        goto cancel;
    }

    hInput = fopen(storage_object->input_path, "rb");
    if (hInput == NULL) {
        LOG_ERR("Cannot open file %s", storage_object->input_path);
        storage_object->progress_state = IOTCS_FAILED;
        return IOTCS_RESULT_INVALID_ARGUMENT;
    }
    storage_object->progress_state = IOTCS_IN_PROGRESS;

    LOG_INFOS(">>>>>>>>>>>>>>>>>>>>> Upload file to SCS");
    GOTO_ERR(iotcs_port_storage_ssl_connect() != IOTCS_RESULT_OK);

    /*write header*/
    if (IS_CLASSIC_OBJECT_STORAGE) {
        /*API header*/
        if (storage_object->endpoint_id && !DISABLE_STORAGE_OBJECT_PREFIX) {
            GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer, IOTCS_RESPONSE_BUFFER_LENGTH, "%s /%s/%s/%s HTTP/1.1\r\n",
                    protocol_request_methods_to_string(PROTOCOL_REQUEST_METHOD_PUT), g_storage_rest_api, storage_object->endpoint_id, storage_object->name)), REQUEST_STRING_LENGTH_ERR_MSG);
        } else {
            GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer, IOTCS_RESPONSE_BUFFER_LENGTH, "%s /%s/%s HTTP/1.1\r\n",
                    protocol_request_methods_to_string(PROTOCOL_REQUEST_METHOD_PUT), g_storage_rest_api, storage_object->name)), REQUEST_STRING_LENGTH_ERR_MSG);
        }
    } else {
        /*API header*/
        GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer, IOTCS_RESPONSE_BUFFER_LENGTH, "%s /%s HTTP/1.1\r\n",
                protocol_request_methods_to_string(PROTOCOL_REQUEST_METHOD_PUT), g_storage_rest_api)), REQUEST_STRING_LENGTH_ERR_MSG);
    }
    cx += written_bytes;
    generate_uri(storage_object);


    if (IS_CLASSIC_OBJECT_STORAGE) {
        /*Host*/
        GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer + cx, IOTCS_RESPONSE_BUFFER_LENGTH - cx,
                "Host: %s\r\n", g_storage_host)), REQUEST_STRING_LENGTH_ERR_MSG);
        cx += written_bytes;
        fseek(hInput, 0L, SEEK_END);
        /*Content-Length*/
        GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer + cx, IOTCS_RESPONSE_BUFFER_LENGTH - cx,
               "Content-length: %d\r\n", (int)ftell(hInput))), REQUEST_STRING_LENGTH_ERR_MSG);
        cx += written_bytes;
        fseek(hInput, 0L, SEEK_SET);
        /*X-Auth-Token*/
        GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer + cx, IOTCS_RESPONSE_BUFFER_LENGTH - cx,
                "X-Auth-Token: %s\r\n", g_storage_auth_token)), REQUEST_STRING_LENGTH_ERR_MSG);
        cx += written_bytes;
        /*Connection*/
        GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer + cx, IOTCS_RESPONSE_BUFFER_LENGTH - cx,
                "Connection: close\r\n")), REQUEST_STRING_LENGTH_ERR_MSG);
        cx += written_bytes;
    } else {
        /*Connection*/
        GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer + cx, IOTCS_RESPONSE_BUFFER_LENGTH - cx,
                "Accept-Encoding: identity\r\n")), REQUEST_STRING_LENGTH_ERR_MSG);
        cx += written_bytes;
        /*Connection*/
        GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer + cx, IOTCS_RESPONSE_BUFFER_LENGTH - cx,
                "Connection: close\r\n")), REQUEST_STRING_LENGTH_ERR_MSG);
        cx += written_bytes;
        fseek(hInput, 0L, SEEK_END);
        /*Content-Length*/
        GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer + cx, IOTCS_RESPONSE_BUFFER_LENGTH - cx,
               "Content-length: %d\r\n", (int)ftell(hInput))), REQUEST_STRING_LENGTH_ERR_MSG);
        cx += written_bytes;
        fseek(hInput, 0L, SEEK_SET);
        for (i = 0; i < g_storage_auth_header_len; i++) {
            GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer + cx, IOTCS_RESPONSE_BUFFER_LENGTH - cx,
                    "%s: %s\r\n", g_storage_auth_header[i].key, g_storage_auth_header[i].value)), REQUEST_STRING_LENGTH_ERR_MSG);
            cx += written_bytes;
        }
    }
    if (storage_object->metadata) {
        iotcs_named_value *cur = storage_object->metadata;
        while (cur) {
            if (IS_CLASSIC_OBJECT_STORAGE) {
                GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer + cx, IOTCS_RESPONSE_BUFFER_LENGTH - cx,
                        "X-Object-Meta-%s: %s\r\n", cur->name, cur->typed_value.value.string_value)), REQUEST_STRING_LENGTH_ERR_MSG);
            } else {
                GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer + cx, IOTCS_RESPONSE_BUFFER_LENGTH - cx,
                        "opc-meta-%s: %s\r\n", cur->name, cur->typed_value.value.string_value)), REQUEST_STRING_LENGTH_ERR_MSG);
            }
            cx += written_bytes;
            cur = cur->next;
        }
    }
    /*Content-Type*/
    GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer + cx, IOTCS_RESPONSE_BUFFER_LENGTH - cx,
            "content-type: %s\r\n\r\n", storage_object->content_type)), REQUEST_STRING_LENGTH_ERR_MSG);
    cx += written_bytes;

    LOG_INFOSN(g_storage_response_buffer, cx);
    GOTO_ERR(iotcs_port_storage_ssl_write(g_storage_response_buffer, cx) != IOTCS_RESULT_OK);

    char* chunk_buffer = &g_storage_response_buffer[0];
    char* tmp_buffer = &g_storage_response_buffer[IOTCS_RESPONSE_BUFFER_LENGTH / 2 + CHUNK_OVERHEAD + 1]; //NULL terminated symbols

    /*write chunked data*/
    while (0 < (written_bytes = fread(tmp_buffer, 1, IOTCS_RESPONSE_BUFFER_LENGTH / 2 - CHUNK_OVERHEAD - 2, hInput))) {
        int tmp_bytes;
        tmp_buffer[written_bytes] = '\0'; //add null terminated symbol

        GOTO_ERR(iotcs_port_storage_ssl_write(tmp_buffer, written_bytes) != IOTCS_RESULT_OK);
        bytes_total += written_bytes;

        if (cl_storage_progress_cb_h(request->storage,
                storage_object->progress_state, bytes_total, NULL) == IOTCS_TRUE) {
            goto cancel;
        }
    }

    /*read the response*/
    GOTO_ERR(iotcs_port_storage_ssl_read(g_storage_response_buffer, IOTCS_RESPONSE_BUFFER_LENGTH, &written_bytes) != IOTCS_RESULT_OK);
    LOG_INFOSN(g_storage_response_buffer, written_bytes);

    /*disconnect the socket*/
    GOTO_ERR(iotcs_port_storage_ssl_disconnect() != IOTCS_RESULT_OK);
    fclose(hInput);
    hInput = NULL;

    if (!IS_CLASSIC_OBJECT_STORAGE) {
        tmp = strstr(g_storage_response_buffer + 1, "opc-content-md5:");
        if (tmp != NULL) {
            tmp += strlen("opc-content-md5:");
            while (*tmp == ' ') {
                tmp++;
            }
            // Copy base64 encoded hash
            int base64hash_len = 0;
            size_t hash_len = IOTCS_MD5_DIGEST_LENGTH;
            size_t tmp_hash_len = TMP_BUFFER_LEN;
            while (*tmp != '\r' && *tmp != '\0') {
                hash[base64hash_len] = *tmp;
                tmp++;
                base64hash_len++;
            }
            hash[base64hash_len] = 0;

            if (util_file_get_md5(calc_hash, storage_object->input_path,
                                tmp_hash_buffer, TMP_BUFFER_LEN) == IOTCS_RESULT_OK) {
                calc_hash[IOTCS_MD5_DIGEST_LENGTH] = '\0';
                if (iotcs_port_crypto_encode_base64(tmp_hash_buffer, &tmp_hash_len,
                            calc_hash, hash_len) == IOTCS_RESULT_OK) {
                    tmp_hash_buffer[tmp_hash_len] = '\0';
                }
                if (strncmp(hash, tmp_hash_buffer, USEFUL_DIGEST_LEN) != 0) {
                    LOG_ERR("Hash doesn't match %s != %s", hash, tmp_hash_buffer);
                    storage_object->progress_state = IOTCS_FAILED;
                    result = IOTCS_RESULT_FAIL;
                }
            }
        }
    } else {
        tmp = strstr(g_storage_response_buffer + 1, "ETag:");
        if (tmp != NULL) {
            tmp += strlen("ETag:");
            while (*tmp == ' ') {
                tmp++;
            }
            // Copy base64 encoded hash
            int hash_len = 0;
            while (*tmp != '\r' && *tmp != '\0') {
                hash[hash_len] = *tmp;
                tmp++;
                hash_len++;
            }
            hash[hash_len] = 0;

            if (util_file_get_md5(calc_hash, storage_object->input_path,
                                tmp_hash_buffer, TMP_BUFFER_LEN) == IOTCS_RESULT_OK) {
                util_digest_to_hex(tmp_hash_buffer, (unsigned char*)calc_hash, IOTCS_MD5_DIGEST_LENGTH);
                if (strcmp(hash, tmp_hash_buffer) != 0) {
                    LOG_ERR("Hash doesn't match %s != %s", hash, calc_hash);
                    storage_object->progress_state = IOTCS_FAILED;
                    result = IOTCS_RESULT_FAIL;
                }
            }
        }
    }

    /*parse the response buffer*/
    GOTO_ERR(http_parse(response, g_storage_response_buffer, written_bytes) != IOTCS_RESULT_OK);

    LOG_INFOSN(g_storage_response_buffer, written_bytes);
    tmp = strstr(g_storage_response_buffer, "X-Last-Modified-Timestamp:");
    if (tmp != NULL) {
        tmp += strlen("X-Last-Modified-Timestamp:");
        while (*tmp == ' ') {
            tmp++;
        }
        int last_modify = atoi(tmp);
        storage_object->content_date = IOTCS_SEC_TO_MILLISEC(last_modify);
    }

    if (storage_object->progress_state == IOTCS_IN_PROGRESS) {
        result = IOTCS_RESULT_OK;
        if (response->status_code == PROTOCOL_RESPONSE_CODE_CREATED ||
            response->status_code == PROTOCOL_RESPONSE_CODE_OK) {
            storage_object->progress_state = IOTCS_COMPLETED;
        } else {
            storage_object->progress_state = IOTCS_FAILED;
            result = IOTCS_RESULT_FAIL;
        }
    }

    return result;
error:
    storage_object->progress_state = IOTCS_FAILED;
    if (hInput) {
        fclose(hInput);
    }
    iotcs_port_storage_ssl_disconnect();
    return result;

cancel:
    storage_object->progress_state = IOTCS_CANCELED;
    if (hInput) {
        fclose(hInput);
    }
    iotcs_port_storage_ssl_disconnect();
    return IOTCS_RESULT_OPERATION_CANCELED;
}

struct get_next_string_args {
    FILE* hInput;
    char* tmp_buffer;
    int tmp_buffer_len;
};

static const char* get_next_line(void *fptr, int *size) {
    struct get_next_string_args* data = (struct get_next_string_args*) fptr;
    int result = 0, read_bytes = 0;
    /*write chunked data*/
    if ((read_bytes = fread(data->tmp_buffer, 1, data->tmp_buffer_len, data->hInput)) > 0) {
        *size = read_bytes;
        return data->tmp_buffer;
    } else {
        return NULL;
    }

}

iotcs_result util_file_get_md5(char *output_buffer, const char *path, char* tmp_buffer, int tmp_buffer_len) {
    FILE* hInput = fopen(path, "rb");
    if (hInput != NULL) {
        struct get_next_string_args ctx = {hInput, tmp_buffer, tmp_buffer_len};
        iotcs_result rv;
        if (IOTCS_RESULT_OK != (rv = iotcs_port_md5_file((unsigned char*)output_buffer, get_next_line, (void*)&ctx))) {
            return rv;
        }

        fclose(hInput);
        hInput = NULL;
    } else {
        LOG_ERR("Cannot open file %s to calculate hash", path);
        return IOTCS_RESULT_FAIL;
    }
    return IOTCS_RESULT_OK;
}

/*public/safe*/
iotcs_result cl_internal_storage_file_upload(iotcs_storage_request* request) {
    iotcs_result result;
    protocol_response response;
    struct sd_storage_object_t *storage_object = (struct sd_storage_object_t *)request->storage->object;
    char *object_path = build_object_path(storage_object);

    UTIL_LOCK_STORAGE();
    if (IOTCS_RESULT_OK != storage_authentication(object_path, "PUT")) {
        util_free(object_path);
        UTIL_UNLOCK_STORAGE();
        storage_object->progress_state = IOTCS_FAILED;
        return IOTCS_RESULT_CANNOT_AUTHORIZE;
    }
    util_free(object_path);

    result = cl_internal_send_upload_request_unsafe(request, &response);
    switch (response.status_code) {
        case PROTOCOL_RESPONSE_CODE_OK:
        case PROTOCOL_RESPONSE_CODE_CREATED:
            break;
        case PROTOCOL_RESPONSE_CODE_UNAUTHORIZED:
            g_storage_reconfig_required = IOTCS_TRUE;
            object_path = build_object_path(storage_object);
            if (IOTCS_RESULT_OK != storage_authentication(object_path, "PUT")) {
                util_free(object_path);
                storage_object->progress_state = IOTCS_FAILED;
                LOG_ERRS("scs_reconfig_unsafe method was failed");
                result = IOTCS_RESULT_CANNOT_AUTHORIZE;
                cl_storage_progress_cb_h(request->storage, storage_object->progress_state, 0, NULL);
                UTIL_UNLOCK_STORAGE();
                return result;
            }
            util_free(object_path);

            // Retry
            result = cl_internal_send_upload_request_unsafe(request, &response);
            if (result != IOTCS_RESULT_OK) {
                storage_object->progress_state = IOTCS_FAILED;
                cl_storage_progress_cb_h(request->storage, storage_object->progress_state, 0, NULL);
            }
            break;
        default:
            cl_storage_progress_cb_h(request->storage, storage_object->progress_state, 0, NULL);
            result = IOTCS_RESULT_FAIL;
    }

    UTIL_UNLOCK_STORAGE();
    return result;
}

/*Should be guarded (or called from guarded methods) by UTIL_LOCK_STORAGE/UTIL_UNLOCK_STORAGE*/
iotcs_result cl_internal_send_download_request_unsafe(iotcs_storage_request* request, protocol_response* response) {
    iotcs_result result;
    int written_bytes = 0;
    int cx = 0, i = 0;
    char* tmp = NULL;
    char* header_str = NULL;
    char* body_str = NULL;
    int64_t bytes_total = 0;
    FILE* hInput = NULL;
    char calc_hash[IOTCS_MD5_DIGEST_LENGTH + 1], hash[IOTCS_MD5_DIGEST_LENGTH + 1];
    char tmp_hash_buffer[TMP_BUFFER_LEN];
    struct sd_storage_object_t *storage_object = (struct sd_storage_object_t *)request->storage->object;

    /*check cancellation mark*/
    if (cl_storage_progress_cb_h(request->storage, storage_object->progress_state, bytes_total, NULL) == IOTCS_TRUE) {
        goto cancel;
    }
    response->status_code = PROTOCOL_RESPONSE_CODE_UNKNOWN_CODE;

    hInput = fopen(storage_object->output_path, "wb");
    if (hInput == NULL) {
        LOG_ERR("Cannot open file %s", storage_object->output_path);
        storage_object->progress_state = IOTCS_FAILED;
        return IOTCS_RESULT_INVALID_ARGUMENT;
    }

    LOG_INFOS(">>>>>>>>>>>>>>>>>>>>> Download file to SCS");
    GOTO_ERR(iotcs_port_storage_ssl_connect() != IOTCS_RESULT_OK);
    storage_object->progress_state = IOTCS_IN_PROGRESS;

    /*write header*/
    if (IS_CLASSIC_OBJECT_STORAGE) {
        /*API header*/
        if (storage_object->endpoint_id && !DISABLE_STORAGE_OBJECT_PREFIX) {
            GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer, IOTCS_RESPONSE_BUFFER_LENGTH, "%s /%s/%s/%s HTTP/1.1\r\n",
                    protocol_request_methods_to_string(PROTOCOL_REQUEST_METHOD_GET), g_storage_rest_api, storage_object->endpoint_id, storage_object->name)), REQUEST_STRING_LENGTH_ERR_MSG);
        } else {
            GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer, IOTCS_RESPONSE_BUFFER_LENGTH, "%s /%s/%s HTTP/1.1\r\n",
                    protocol_request_methods_to_string(PROTOCOL_REQUEST_METHOD_GET), g_storage_rest_api, storage_object->name)), REQUEST_STRING_LENGTH_ERR_MSG);
        }
    } else {
        /*API header*/
        GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer, IOTCS_RESPONSE_BUFFER_LENGTH, "%s /%s HTTP/1.1\r\n",
                protocol_request_methods_to_string(PROTOCOL_REQUEST_METHOD_GET), g_storage_rest_api)), REQUEST_STRING_LENGTH_ERR_MSG);
    }
    cx += written_bytes;


    /*Connection*/
    GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer + cx, IOTCS_RESPONSE_BUFFER_LENGTH - cx,
            "Connection: close\r\n")), REQUEST_STRING_LENGTH_ERR_MSG);
    cx += written_bytes;
    if (IS_CLASSIC_OBJECT_STORAGE) {
        /*Host*/
        GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer + cx, IOTCS_RESPONSE_BUFFER_LENGTH - cx,
                "Host: %s\r\n", g_storage_host)), REQUEST_STRING_LENGTH_ERR_MSG);
        cx += written_bytes;
        /*X-Auth-Token*/
        GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer + cx, IOTCS_RESPONSE_BUFFER_LENGTH - cx,
                "X-Auth-Token: %s\r\n", g_storage_auth_token)), REQUEST_STRING_LENGTH_ERR_MSG);
        cx += written_bytes;
    } else {
        for (i = 0; i < g_storage_auth_header_len; i++) {
            GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer + cx, IOTCS_RESPONSE_BUFFER_LENGTH - cx,
                    "%s: %s\r\n", g_storage_auth_header[i].key, g_storage_auth_header[i].value)), REQUEST_STRING_LENGTH_ERR_MSG);
            cx += written_bytes;
        }
    }

    /*Connection*/
    GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer + cx, IOTCS_RESPONSE_BUFFER_LENGTH - cx,
            "\r\n")), REQUEST_STRING_LENGTH_ERR_MSG);
    cx += written_bytes;

    LOG_INFOSN(g_storage_response_buffer, cx);
    GOTO_ERR(iotcs_port_storage_ssl_write(g_storage_response_buffer, cx) != IOTCS_RESULT_OK);

    /*read first part*/
    GOTO_ERR(iotcs_port_storage_ssl_read(g_storage_response_buffer, IOTCS_RESPONSE_BUFFER_LENGTH, &written_bytes) != IOTCS_RESULT_OK);

    /*modify the response buffer to find headers*/
    GOTO_ERR_MSG(NULL == (tmp = strstr(g_storage_response_buffer, "\r\n\r\n")), "Bad format of the http response");
    strncpy(tmp, "\0\0\0\0", 4);

    header_str = &g_storage_response_buffer[0];
    body_str = tmp + 4;
    written_bytes -= strlen(header_str) + 4;
    LOG_INFOS(header_str);

    /*prepare the protocol_response*/
    response->body = NULL;
    response->min_accept_bytes = 0;

    LOG_INFOSN(g_storage_response_buffer, written_bytes);
    /*parse response code*/
    /* Drop "HTTP/1.1 " */
    tmp = &header_str[9];
    int code = atoi(tmp);
    switch (code) {
        case 200:
        case 202:
        case 206:
            response->status_code = PROTOCOL_RESPONSE_CODE_OK;
            break;
        case 400:
            response->status_code = PROTOCOL_RESPONSE_CODE_BAD_REQUEST;
            break;
        case 401:
            response->status_code = PROTOCOL_RESPONSE_CODE_UNAUTHORIZED;
            break;
        case 404:
            response->status_code = PROTOCOL_RESPONSE_CODE_NOT_FOUND;
            break;
        case 409:
            response->status_code = PROTOCOL_RESPONSE_CODE_CONFLICT;
            break;
        case 500:
        default:
            response->status_code = PROTOCOL_RESPONSE_CODE_INTERNAL_SERVER_ERROR;
            break;
    }

    GOTO_ERR(response->status_code != PROTOCOL_RESPONSE_CODE_OK);

    /*TODO: Think how to support chunked data*/
    GOTO_ERR_MSG(NULL != (tmp = strstr(header_str, "chunked")), "Chunked data is unsupported yet");

    if (written_bytes > 0) {
        GOTO_ERR(0 > fwrite(body_str, sizeof (char), written_bytes, hInput));
        bytes_total += written_bytes;
        if (cl_storage_progress_cb_h(request->storage, storage_object->progress_state, bytes_total, NULL) == IOTCS_TRUE) {
            goto cancel;
        }
    }

    do {
        GOTO_ERR(iotcs_port_storage_ssl_read(g_storage_response_buffer, IOTCS_RESPONSE_BUFFER_LENGTH, &written_bytes) != IOTCS_RESULT_OK);
        if (written_bytes > 0) {
            GOTO_ERR(0 > fwrite(g_storage_response_buffer, sizeof (char), written_bytes, hInput));
        }
        bytes_total += written_bytes;
        if (cl_storage_progress_cb_h(request->storage, storage_object->progress_state, bytes_total, NULL) == IOTCS_TRUE) {
            goto cancel;
        }
   } while (written_bytes > 0);

    /*disconnect the socket*/
    GOTO_ERR(iotcs_port_storage_ssl_disconnect() != IOTCS_RESULT_OK);

    fclose(hInput);
    hInput = NULL;
    result = IOTCS_RESULT_OK;
    LOG_INFOS(">>>>>>>>>>>>>Disconnected");
    LOG_INFOSN(g_storage_response_buffer, written_bytes);
    tmp = strstr(g_storage_response_buffer, "X-Last-Modified-Timestamp:");
    if (tmp != NULL) {
        tmp += strlen("X-Last-Modified-Timestamp:");
        while (*tmp == ' ') {
            tmp++;
        }
        int last_modify = atoi(tmp);
        storage_object->content_date = IOTCS_SEC_TO_MILLISEC(last_modify);
    }

    if (!IS_CLASSIC_OBJECT_STORAGE) {
        tmp = strstr(g_storage_response_buffer + 1, "opc-content-md5:");
        if (tmp != NULL) {
            tmp += strlen("opc-content-md5:");
            while (*tmp == ' ') {
                tmp++;
            }
            // Copy base64 encoded hash
            int base64hash_len = 0;
            size_t hash_len = IOTCS_MD5_DIGEST_LENGTH;
            size_t tmp_hash_len = TMP_BUFFER_LEN;
            while (*tmp != '\r' && *tmp != '\0') {
                hash[base64hash_len] = *tmp;
                tmp++;
                base64hash_len++;
            }
            hash[base64hash_len] = 0;

            if (util_file_get_md5(calc_hash, storage_object->input_path,
                                tmp_hash_buffer, TMP_BUFFER_LEN) == IOTCS_RESULT_OK) {
                calc_hash[IOTCS_MD5_DIGEST_LENGTH] = '\0';
                if (iotcs_port_crypto_encode_base64(tmp_hash_buffer, &tmp_hash_len,
                            calc_hash, hash_len) == IOTCS_RESULT_OK) {
                    tmp_hash_buffer[tmp_hash_len] = '\0';
                }
                if (strncmp(hash, tmp_hash_buffer, USEFUL_DIGEST_LEN) != 0) {
                    LOG_ERR("Hash doesn't match %s != %s", hash, tmp_hash_buffer);
                    storage_object->progress_state = IOTCS_FAILED;
                    result = IOTCS_RESULT_FAIL;
                }
            }
        }
    } else {
        tmp = strstr(g_storage_response_buffer + 1, "ETag:");
        if (tmp != NULL) {
            tmp += strlen("ETag:");
            while (*tmp == ' ') {
                tmp++;
            }
            // Copy base64 encoded hash
            int hash_len = 0;
            while (*tmp != '\r' && *tmp != '\0') {
                hash[hash_len] = *tmp;
                tmp++;
                hash_len++;
            }
            hash[hash_len] = 0;

            if (util_file_get_md5(calc_hash, storage_object->input_path,
                                tmp_hash_buffer, TMP_BUFFER_LEN) == IOTCS_RESULT_OK) {
                util_digest_to_hex(tmp_hash_buffer, (unsigned char*)calc_hash, IOTCS_MD5_DIGEST_LENGTH);
                if (strcmp(hash, tmp_hash_buffer) != 0) {
                    LOG_ERR("Hash doesn't match %s != %s", hash, calc_hash);
                    storage_object->progress_state = IOTCS_FAILED;
                    result = IOTCS_RESULT_FAIL;
                }
            }
        }
    }

    if (response->status_code == PROTOCOL_RESPONSE_CODE_OK ||
        response->status_code == PROTOCOL_RESPONSE_CODE_PARTIAL) {
        cl_storage_progress_cb_h(request->storage, storage_object->progress_state, bytes_total, NULL);
        storage_object->progress_state = IOTCS_COMPLETED;
    } else {
        result = IOTCS_RESULT_FAIL;
    }

    return result;
error:
    if (hInput) {
        fclose(hInput);
    }
    iotcs_port_storage_ssl_disconnect();
    if (response->status_code == PROTOCOL_RESPONSE_CODE_UNAUTHORIZED) {
        return IOTCS_RESULT_OK;
    }
    storage_object->progress_state = IOTCS_FAILED;

    return IOTCS_RESULT_FAIL;

cancel:
    storage_object->progress_state = IOTCS_CANCELED;

    if (hInput) {
        fclose(hInput);
    }
    iotcs_port_storage_ssl_disconnect();
    return IOTCS_RESULT_OPERATION_CANCELED;
}

/*public/safe*/
iotcs_result cl_internal_storage_file_download(iotcs_storage_request* request) {
    iotcs_result result;
    protocol_response response;
    struct sd_storage_object_t *storage_object = (struct sd_storage_object_t *)request->storage->object;
    char *object_path = build_object_path(storage_object);
    UTIL_LOCK_STORAGE();

    if (IOTCS_RESULT_OK != storage_authentication(object_path, "GET")) {
        util_free(object_path);
        storage_object->progress_state = IOTCS_FAILED;
        UTIL_UNLOCK_STORAGE();
        return IOTCS_RESULT_CANNOT_AUTHORIZE;
    }
    util_free(object_path);

    result = cl_internal_send_download_request_unsafe(request, &response);

    switch (response.status_code) {
        case PROTOCOL_RESPONSE_CODE_OK:
        case PROTOCOL_RESPONSE_CODE_PARTIAL:
            break;
        case PROTOCOL_RESPONSE_CODE_UNAUTHORIZED:
            g_storage_reconfig_required = IOTCS_TRUE;
            object_path = build_object_path(storage_object);
            if (IOTCS_RESULT_OK != storage_authentication(object_path, "GET")) {
                util_free(object_path);
                LOG_ERRS("scs_reconfig_unsafe method was failed");
                storage_object->progress_state = IOTCS_FAILED;
                result = IOTCS_RESULT_CANNOT_AUTHORIZE;
                cl_storage_progress_cb_h(request->storage, storage_object->progress_state, 0, NULL);
                UTIL_UNLOCK_STORAGE();
                return result;
            }
            util_free(object_path);

            // Retry
            result = cl_internal_send_download_request_unsafe(request, &response);
            if (result != IOTCS_RESULT_OK) {
                storage_object->progress_state = IOTCS_FAILED;
                cl_storage_progress_cb_h(request->storage, storage_object->progress_state, 0, NULL);
            }
            break;
        default:
            cl_storage_progress_cb_h(request->storage, storage_object->progress_state, 0, NULL);
            result = IOTCS_RESULT_FAIL;
    }
    UTIL_UNLOCK_STORAGE();

    return result;
}

/*Should be guarded (or called from guarded methods) by UTIL_LOCK_STORAGE/UTIL_UNLOCK_STORAGE*/
static iotcs_result send_file_information_request_unsafe(struct iotcs_uri_object_t *locator, protocol_response* response) {
    int written_bytes = 0;
    int cx = 0, i = 0;
    char *tmp = NULL;
    struct sd_storage_object_t *storage_object = (struct sd_storage_object_t *)locator->object;

    LOG_INFOS(">>>>>>>>>>>>>>>>>>>>> Download file info from SCS");
    GOTO_ERR(iotcs_port_storage_ssl_connect() != IOTCS_RESULT_OK);

    /*write header*/

    /*API header*/
    if (IS_CLASSIC_OBJECT_STORAGE) {
        /*API header*/
        if (storage_object->endpoint_id && !DISABLE_STORAGE_OBJECT_PREFIX) {
            GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer, IOTCS_RESPONSE_BUFFER_LENGTH, "%s /%s/%s/%s HTTP/1.1\r\n",
                    protocol_request_methods_to_string(PROTOCOL_REQUEST_METHOD_HEAD), g_storage_rest_api, storage_object->endpoint_id, storage_object->name)), REQUEST_STRING_LENGTH_ERR_MSG);
        } else {
            GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer, IOTCS_RESPONSE_BUFFER_LENGTH, "%s /%s/%s HTTP/1.1\r\n",
                    protocol_request_methods_to_string(PROTOCOL_REQUEST_METHOD_HEAD), g_storage_rest_api, storage_object->name)), REQUEST_STRING_LENGTH_ERR_MSG);
        }
    } else {
        /*API header*/
        GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer, IOTCS_RESPONSE_BUFFER_LENGTH, "%s /%s HTTP/1.1\r\n",
                protocol_request_methods_to_string(PROTOCOL_REQUEST_METHOD_HEAD), g_storage_rest_api)), REQUEST_STRING_LENGTH_ERR_MSG);
    }
    cx += written_bytes;

    /*Connection*/
    GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer + cx, IOTCS_RESPONSE_BUFFER_LENGTH - cx,
            "Connection: close\r\n")), REQUEST_STRING_LENGTH_ERR_MSG);
    cx += written_bytes;
    if (IS_CLASSIC_OBJECT_STORAGE) {
        /*Host*/
        GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer + cx, IOTCS_RESPONSE_BUFFER_LENGTH - cx,
                "Host: %s\r\n", g_storage_host)), REQUEST_STRING_LENGTH_ERR_MSG);
        cx += written_bytes;
        /*X-Auth-Token*/
        GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer + cx, IOTCS_RESPONSE_BUFFER_LENGTH - cx,
                "X-Auth-Token: %s\r\n", g_storage_auth_token)), REQUEST_STRING_LENGTH_ERR_MSG);
        cx += written_bytes;
    } else {
        for (i = 0; i < g_storage_auth_header_len; i++) {
            GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer + cx, IOTCS_RESPONSE_BUFFER_LENGTH - cx,
                    "%s: %s\r\n", g_storage_auth_header[i].key, g_storage_auth_header[i].value)), REQUEST_STRING_LENGTH_ERR_MSG);
            cx += written_bytes;
        }
    }
    /*Connection*/
    GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer + cx, IOTCS_RESPONSE_BUFFER_LENGTH - cx,
            "\r\n")), REQUEST_STRING_LENGTH_ERR_MSG);
    cx += written_bytes;

    LOG_INFOSN(g_storage_response_buffer, cx);
    GOTO_ERR(iotcs_port_storage_ssl_write(g_storage_response_buffer, cx) != IOTCS_RESULT_OK);

    /*read first part*/
    GOTO_ERR(iotcs_port_storage_ssl_read(g_storage_response_buffer, IOTCS_RESPONSE_BUFFER_LENGTH, &written_bytes) != IOTCS_RESULT_OK);
    LOG_INFOSN(g_storage_response_buffer, written_bytes);

    /*disconnect the socket*/
    GOTO_ERR(iotcs_port_storage_ssl_disconnect() != IOTCS_RESULT_OK);

    tmp = strstr(g_storage_response_buffer, "X-Last-Modified-Timestamp:");
    if (tmp != NULL) {
        tmp += strlen("X-Last-Modified-Timestamp:");
        while (*tmp == ' ') {
            tmp++;
        }
        int last_modify = atoi(tmp);
        storage_object->content_date = IOTCS_SEC_TO_MILLISEC(last_modify);
    }

    if (IS_CLASSIC_OBJECT_STORAGE) {
        tmp = strstr(g_storage_response_buffer, "X-Object-Meta-");
    } else {
        tmp = strstr(g_storage_response_buffer, "opc-meta-");
    }
    while (tmp != NULL) {
        if (IS_CLASSIC_OBJECT_STORAGE) {
            tmp += strlen("X-Object-Meta-");
        } else {
            tmp += strlen("opc-meta-");
        }
        char *name, *value;
        int len = 0;
        while (*(tmp+len) != ':') {
            len++;
        }
        name = util_safe_strncpy(tmp, len);
        len++;
        tmp += len;
        len = 0;
        while (*(tmp) == ' ') {
            tmp++;
        }
        while (isalnum(*(tmp+len))) {
            len++;
        }
        value = util_safe_strncpy(tmp, len);
        iotcs_storage_object_set_custom_metadata(locator, name, value);
        util_free(name);
        util_free(value);
        if (IS_CLASSIC_OBJECT_STORAGE) {
            tmp = strstr(g_storage_response_buffer, "X-Object-Meta-");
        } else {
            tmp = strstr(g_storage_response_buffer, "opc-meta-");
        }
    }
    /*parse the response buffer*/
    GOTO_ERR(http_parse(response, g_storage_response_buffer, written_bytes) != IOTCS_RESULT_OK);

    return IOTCS_RESULT_OK;
error:
    iotcs_port_storage_ssl_disconnect();
    return IOTCS_RESULT_FAIL;
}

void generate_uri(struct sd_storage_object_t *storage_object) {
    if (g_storage_container_url) {
        int expected_size = strlen(g_storage_container_url)+strlen(storage_object->name)+2;
        if (storage_object->endpoint_id && !DISABLE_STORAGE_OBJECT_PREFIX) {
            expected_size += strlen(storage_object->endpoint_id) + 1;
        }
        char *tmp_buffer = (char*)util_calloc(expected_size, sizeof(char));
        if (tmp_buffer) {
            if (IS_CLASSIC_OBJECT_STORAGE) {
                if (storage_object->endpoint_id && !DISABLE_STORAGE_OBJECT_PREFIX) {
                    util_safe_snprintf(tmp_buffer, expected_size, "%s/%s/%s", g_storage_container_url, storage_object->endpoint_id, storage_object->name);
                } else {
                    util_safe_snprintf(tmp_buffer, expected_size, "%s/%s", g_storage_container_url, storage_object->name);
                }
            } else {
                util_safe_snprintf(tmp_buffer, expected_size, g_storage_container_url);
            }
            storage_object->uri = tmp_buffer;
        }
    }
}

/*public/safe*/
iotcs_result iotcs_storage_object_fill_file_info(struct iotcs_uri_object_t *locator) {
    iotcs_result result;
    protocol_response response;
    struct sd_storage_object_t *storage_object = (struct sd_storage_object_t *)locator->object;
    char *object_path = build_object_path(storage_object);

    UTIL_LOCK_STORAGE();
    if (IOTCS_RESULT_OK != storage_authentication(object_path, "HEAD")) {
        util_free(object_path);
        UTIL_UNLOCK_STORAGE();
        return IOTCS_RESULT_CANNOT_AUTHORIZE;
    }
    util_free(object_path);

    result = send_file_information_request_unsafe(locator, &response);

    switch (response.status_code) {
        case PROTOCOL_RESPONSE_CODE_ACCEPTED:
        case PROTOCOL_RESPONSE_CODE_OK:
        case PROTOCOL_RESPONSE_CODE_CREATED:
            break;
        case PROTOCOL_RESPONSE_CODE_UNAUTHORIZED:
            g_storage_reconfig_required = IOTCS_TRUE;
            object_path = build_object_path(storage_object);
            if (IOTCS_RESULT_OK != storage_authentication(object_path, "HEAD")) {
                util_free(object_path);
                UTIL_UNLOCK_STORAGE();
                LOG_ERRS("scs_reconfig_unsafe method was failed");
                return IOTCS_RESULT_CANNOT_AUTHORIZE;
            }
            util_free(object_path);
            result = send_file_information_request_unsafe(locator, &response);
            break;
        default:
            UTIL_UNLOCK_STORAGE();
            return IOTCS_RESULT_FAIL;
    }

    if (result == IOTCS_RESULT_OK &&
            (response.status_code == PROTOCOL_RESPONSE_CODE_ACCEPTED ||
            response.status_code == PROTOCOL_RESPONSE_CODE_OK ||
            response.status_code == PROTOCOL_RESPONSE_CODE_CREATED)) {
        /*update locator*/
        storage_object->content_length = atoi(response.content_length);
        storage_object->content_type = strdup(response.content_type);
        generate_uri(storage_object);
        UTIL_UNLOCK_STORAGE();
        return IOTCS_RESULT_OK;
    }
    UTIL_UNLOCK_STORAGE();

    return result;
}

/*public*/
iotcs_result iotcs_storage_dispatcher_queue(struct iotcs_uri_object_t *object) {
    iotcs_storage_request* request = (iotcs_storage_request*)util_calloc(1, sizeof(iotcs_storage_request));
    struct sd_storage_object_t *storage_object = (struct sd_storage_object_t *)object->object;
    if (storage_object->progress_state == IOTCS_COMPLETED) {
        return IOTCS_RESULT_OK;
    }
    
    if (!storage_object->input_path && !storage_object->output_path && IS_EXTERNAL_OBJECT(object)) {
        return IOTCS_RESULT_INVALID_ARGUMENT;
    }

    generate_uri(storage_object);
    if (storage_object->progress_state == IOTCS_IN_PROGRESS || storage_object->progress_state == IOTCS_QUEUED) {
        return IOTCS_RESULT_INVALID_ARGUMENT;
    }
    storage_object->progress_state = IOTCS_QUEUED;
    request->storage = object;
    object->ref_cnt++;
    iotcs_result rv = cl_internal_storage_dispatcher_queue_safe(request);
    if (rv != IOTCS_RESULT_OK) {
        storage_object->progress_state = IOTCS_FAILED;
        util_free(request);
    }
    return rv;
}

void iotcs_storage_dispatcher_set_callback(iotcs_storage_dispatcher_progress_callback callback) {
    if (callback == NULL) {
        progress_callback = cl_storage_progress_cb;
    } else {
        progress_callback = callback;
    }
}

/*only for MD auth - should be replaced to more smart method that used default library MD in correct way*/
iotcs_result cl_internal_send_reconfig_request_md() {
    protocol_response response;
    int written_bytes;
    int cx = 0;

    LOG_INFOS(">>>>>>>>>>>>>>>>>>>>> Auth request from SCS dispatcher");

    iotcs_port_storage_ssl_finalize();
    // Only ssl here
    GOTO_ERR(iotcs_port_storage_ssl_init(tam_get_server_host(), tam_get_server_port(), IOTCS_TRUE) != IOTCS_RESULT_OK);
    GOTO_ERR(iotcs_port_storage_ssl_connect() != IOTCS_RESULT_OK);

    /*write header*/

    GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer, IOTCS_RESPONSE_BUFFER_LENGTH, "GET /iot/api/v2/provisioner/storage HTTP/1.1\r\n")),
            REQUEST_STRING_LENGTH_ERR_MSG);
    cx += written_bytes;

    GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer + cx, IOTCS_RESPONSE_BUFFER_LENGTH - cx,
            "Host: %s\r\n", tam_get_server_host())), REQUEST_STRING_LENGTH_ERR_MSG);
    cx += written_bytes;

    GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer + cx, IOTCS_RESPONSE_BUFFER_LENGTH - cx,
            "Authorization: %s\r\n", cl_get_access_token_ptr()->value)), REQUEST_STRING_LENGTH_ERR_MSG);
    cx += written_bytes;

    GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer + cx, IOTCS_RESPONSE_BUFFER_LENGTH - cx,
            "X-EndpointId: %s\r\n\r\n", tam_get_endpoint_id())), REQUEST_STRING_LENGTH_ERR_MSG);
    cx += written_bytes;

    GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer + cx, IOTCS_RESPONSE_BUFFER_LENGTH - cx,
            "Connection: close\r\n")), REQUEST_STRING_LENGTH_ERR_MSG);
    cx += written_bytes;

    GOTO_ERR_MSG(0 > (written_bytes = util_safe_snprintf(g_storage_response_buffer + cx, IOTCS_RESPONSE_BUFFER_LENGTH - cx,
            "Content-Type: application/json\r\n\r\n")), REQUEST_STRING_LENGTH_ERR_MSG);
    cx += written_bytes;

    LOG_INFOSN(g_storage_response_buffer, cx);
    GOTO_ERR(iotcs_port_storage_ssl_write(g_storage_response_buffer, cx) != IOTCS_RESULT_OK);


    /*read the response*/
    GOTO_ERR(iotcs_port_storage_ssl_read(g_storage_response_buffer, IOTCS_RESPONSE_BUFFER_LENGTH, &written_bytes) != IOTCS_RESULT_OK);
    LOG_INFOSN(g_storage_response_buffer, written_bytes);

    /*disconnect the socket*/
    GOTO_ERR(iotcs_port_storage_ssl_disconnect() != IOTCS_RESULT_OK);

    /*parse the response buffer*/
    GOTO_ERR(http_parse(&response, g_storage_response_buffer, written_bytes) != IOTCS_RESULT_OK);

    GOTO_ERR(response.body != NULL);
    GOTO_ERR(scs_parse_auth_response(response.body) != IOTCS_RESULT_OK);

    iotcs_port_storage_ssl_disconnect();
    iotcs_port_storage_ssl_finalize();
    iotcs_bool is_ssl = (strncmp("https", g_storage_container_url, strlen("https")) == 0) ? IOTCS_TRUE : IOTCS_FALSE;
    GOTO_ERR(iotcs_port_storage_ssl_init(g_storage_host, g_storage_port, is_ssl) != IOTCS_RESULT_OK);

    return IOTCS_RESULT_OK;
error:

    iotcs_port_storage_ssl_disconnect();
    return IOTCS_RESULT_FAIL;
}
